@@ -19,8 +19,9 @@
  * 02110-1301, USA
  */
 
-#include <int.h>
-#include <hash-pjw-bare.h>
+#include <limits.h> // WORD_BIT
+
+#include "int.h"
 #include "parser_aux.h"
 #include "gstr.h"
 #include "structure.h"
@@ -28,56 +29,83 @@
 
 char _asn1_identifierMissing[ASN1_MAX_NAME_SIZE + 1];	/* identifier name not found */
 
-/***********************************************/
-/* Type: list_type                             */
-/* Description: type used in the list during   */
-/* the structure creation.                     */
-/***********************************************/
-typedef struct list_struct
+/* Return a hash of the N bytes of X using the method described by
+   Bruno Haible in https://www.haible.de/bruno/hashfunc.html.
+   Note that while many hash functions reduce their result via modulo
+   to a 0..table_size-1 range, this function does not do that.
+
+   This implementation has been changed from size_t -> unsigned int. */
+
+#ifdef __clang__
+__attribute__((no_sanitize("integer")))
+#endif
+_GL_ATTRIBUTE_PURE
+static unsigned int
+_asn1_hash_name (const char *x)
 {
-  asn1_node node;
-  struct list_struct *next;
-} list_type;
+  const unsigned char *s = (unsigned char *) x;
+  unsigned h = 0;
 
+  while (*s)
+    h = (*s++) + ((h << 9) | (h >> (WORD_BIT - 9)));
 
-/* Pointer to the first element of the list */
-list_type *firstElement = NULL;
+  return h;
+}
 
 /******************************************************/
 /* Function : _asn1_add_static_node                   */
 /* Description: creates a new NODE_ASN element and    */
-/* puts it in the list pointed by firstElement.       */
+/* puts it in the list pointed by e_list.       */
 /* Parameters:                                        */
+/*   e_list: of type list_type; must be NULL initially */
 /*   type: type of the new element (see ASN1_ETYPE_   */
 /*         and CONST_ constants).                     */
 /* Return: pointer to the new element.                */
 /******************************************************/
 asn1_node
-_asn1_add_static_node (unsigned int type)
+_asn1_add_static_node (list_type **e_list, unsigned int type)
 {
-  list_type *listElement;
+  list_type *p;
   asn1_node punt;
 
   punt = calloc (1, sizeof (struct asn1_node_st));
   if (punt == NULL)
     return NULL;
 
-  listElement = malloc (sizeof (list_type));
-  if (listElement == NULL)
+  p = malloc (sizeof (list_type));
+  if (p == NULL)
     {
       free (punt);
       return NULL;
     }
 
-  listElement->node = punt;
-  listElement->next = firstElement;
-  firstElement = listElement;
+  p->node = punt;
+  p->next = *e_list;
+  *e_list = p;
 
   punt->type = type;
 
   return punt;
 }
 
+static
+int _asn1_add_static_node2 (list_type **e_list, asn1_node node)
+{
+  list_type *p;
+
+  p = malloc (sizeof (list_type));
+  if (p == NULL)
+    {
+      return -1;
+    }
+
+  p->node = node;
+  p->next = *e_list;
+  *e_list = p;
+
+  return 0;
+}
+
 /**
  * asn1_find_node:
  * @pointer: NODE_ASN element pointer.
@@ -91,9 +119,9 @@ _asn1_add_static_node (unsigned int type)
  * Returns: the search result, or %NULL if not found.
  **/
 asn1_node
-asn1_find_node (asn1_node pointer, const char *name)
+asn1_find_node (asn1_node_const pointer, const char *name)
 {
-  asn1_node p;
+  asn1_node_const p;
   char *n_end, n[ASN1_MAX_NAME_SIZE + 1];
   const char *n_start;
   unsigned int nsize;
@@ -128,12 +156,12 @@ asn1_find_node (asn1_node pointer, const char *name)
 	  n_start = n_end;
 	  n_start++;
 
-	  nhash = hash_pjw_bare (n, nsize);
+	  nhash = _asn1_hash_name (n);
 	}
       else
 	{
-	  nsize = _asn1_str_cpy (n, sizeof (n), n_start);
-	  nhash = hash_pjw_bare (n, nsize);
+	  _asn1_str_cpy (n, sizeof (n), n_start);
+	  nhash = _asn1_hash_name (n);
 
 	  n_start = NULL;
 	}
@@ -152,7 +180,7 @@ asn1_find_node (asn1_node pointer, const char *name)
   else
     {				/* *pointer doesn't have a name */
       if (n_start[0] == 0)
-	return p;
+	return (asn1_node) p;
     }
 
   while (n_start)
@@ -169,12 +197,12 @@ asn1_find_node (asn1_node pointer, const char *name)
 	  n_start = n_end;
 	  n_start++;
 
-	  nhash = hash_pjw_bare (n, nsize);
+	  nhash = _asn1_hash_name (n);
 	}
       else
 	{
-	  nsize = _asn1_str_cpy (n, sizeof (n), n_start);
-	  nhash = hash_pjw_bare (n, nsize);
+	  _asn1_str_cpy (n, sizeof (n), n_start);
+	  nhash = _asn1_hash_name (n);
 	  n_start = NULL;
 	}
 
@@ -206,7 +234,7 @@ asn1_find_node (asn1_node pointer, const char *name)
         return NULL;
     }				/* while */
 
-  return p;
+  return (asn1_node) p;
 }
 
 
@@ -380,20 +408,11 @@ _asn1_append_value (asn1_node node, const void *value, unsigned int len)
 asn1_node
 _asn1_set_name (asn1_node node, const char *name)
 {
-  unsigned int nsize;
-
   if (node == NULL)
     return node;
 
-  if (name == NULL)
-    {
-      node->name[0] = 0;
-      node->name_hash = hash_pjw_bare (node->name, 0);
-      return node;
-    }
-
-  nsize = _asn1_str_cpy (node->name, sizeof (node->name), name);
-  node->name_hash = hash_pjw_bare (node->name, nsize);
+  _asn1_str_cpy (node->name, sizeof (node->name), name ? name : "");
+  node->name_hash = _asn1_hash_name (node->name);
 
   return node;
 }
@@ -407,7 +426,7 @@ _asn1_set_name (asn1_node node, const char *name)
 /* Return: pointer to the NODE_ASN element.                       */
 /******************************************************************/
 asn1_node
-_asn1_cpy_name (asn1_node dst, asn1_node src)
+_asn1_cpy_name (asn1_node dst, asn1_node_const src)
 {
   if (dst == NULL)
     return dst;
@@ -415,7 +434,7 @@ _asn1_cpy_name (asn1_node dst, asn1_node src)
   if (src == NULL)
     {
       dst->name[0] = 0;
-      dst->name_hash = hash_pjw_bare (dst->name, 0);
+      dst->name_hash = _asn1_hash_name (dst->name);
       return dst;
     }
 
@@ -454,16 +473,16 @@ _asn1_set_right (asn1_node node, asn1_node right)
 /* Return: pointer to the last element along the right chain.     */
 /******************************************************************/
 asn1_node
-_asn1_get_last_right (asn1_node node)
+_asn1_get_last_right (asn1_node_const node)
 {
-  asn1_node p;
+  asn1_node_const p;
 
   if (node == NULL)
     return NULL;
   p = node;
   while (p->right)
     p = p->right;
-  return p;
+  return (asn1_node) p;
 }
 
 /******************************************************************/
@@ -501,9 +520,9 @@ _asn1_remove_node (asn1_node node, unsigned int flags)
 /* Return: Null if not found.                                     */
 /******************************************************************/
 asn1_node
-_asn1_find_up (asn1_node node)
+_asn1_find_up (asn1_node_const node)
 {
-  asn1_node p;
+  asn1_node_const p;
 
   if (node == NULL)
     return NULL;
@@ -516,21 +535,58 @@ _asn1_find_up (asn1_node node)
   return p->left;
 }
 
+static
+unsigned _asn1_is_up (asn1_node_const up_cand, asn1_node_const down)
+{
+  asn1_node_const d, u;
+
+  if (up_cand == NULL || down == NULL)
+    return 0;
+
+  d = down;
+
+  while ((u = _asn1_find_up(d)) != NULL && u != d)
+    {
+      if (u == up_cand)
+        return 1;
+      d = u;
+    }
+
+  return 0;
+}
+
+/******************************************************************/
+/* Function : _asn1_delete_node_from_list                         */
+/* Description: deletes the list element given                    */
+/******************************************************************/
+void
+_asn1_delete_node_from_list (list_type *list, asn1_node node)
+{
+  list_type *p = list;
+
+  while (p)
+    {
+      if (p->node == node)
+        p->node = NULL;
+      p = p->next;
+    }
+}
+
 /******************************************************************/
 /* Function : _asn1_delete_list                                   */
 /* Description: deletes the list elements (not the elements       */
 /*  pointed by them).                                             */
 /******************************************************************/
 void
-_asn1_delete_list (void)
+_asn1_delete_list (list_type *e_list)
 {
-  list_type *listElement;
+  list_type *p;
 
-  while (firstElement)
+  while (e_list)
     {
-      listElement = firstElement;
-      firstElement = firstElement->next;
-      free (listElement);
+      p = e_list;
+      e_list = e_list->next;
+      free (p);
     }
 }
 
@@ -540,16 +596,16 @@ _asn1_delete_list (void)
 /*  pointed by them.                                              */
 /******************************************************************/
 void
-_asn1_delete_list_and_nodes (void)
+_asn1_delete_list_and_nodes (list_type *e_list)
 {
-  list_type *listElement;
+  list_type *p;
 
-  while (firstElement)
+  while (e_list)
     {
-      listElement = firstElement;
-      firstElement = firstElement->next;
-      _asn1_remove_node (listElement->node, 0);
-      free (listElement);
+      p = e_list;
+      e_list = e_list->next;
+      _asn1_remove_node (p->node, 0);
+      free (p);
     }
 }
 
@@ -660,22 +716,24 @@ _asn1_change_integer_value (asn1_node node)
   return ASN1_SUCCESS;
 }
 
-
+#define MAX_CONSTANTS 1024
 /******************************************************************/
 /* Function : _asn1_expand_object_id                              */
 /* Description: expand the IDs of an OBJECT IDENTIFIER constant.  */
 /* Parameters:                                                    */
+/*   list: root of an object list                                 */
 /*   node: root of an ASN1 element.                               */
 /* Return:                                                        */
-/*   ASN1_ELEMENT_NOT_FOUND if NODE is NULL,                       */
-/*   otherwise ASN1_SUCCESS                                             */
+/*   ASN1_ELEMENT_NOT_FOUND if NODE is NULL,                      */
+/*   otherwise ASN1_SUCCESS                                       */
 /******************************************************************/
 int
-_asn1_expand_object_id (asn1_node node)
+_asn1_expand_object_id (list_type **list, asn1_node node)
 {
   asn1_node p, p2, p3, p4, p5;
   char name_root[ASN1_MAX_NAME_SIZE], name2[2 * ASN1_MAX_NAME_SIZE + 1];
-  int move, tlen;
+  int move, tlen, tries;
+  unsigned max_constants;
 
   if (node == NULL)
     return ASN1_ELEMENT_NOT_FOUND;
@@ -684,6 +742,7 @@ _asn1_expand_object_id (asn1_node node)
 
   p = node;
   move = DOWN;
+  tries = 0;
 
   while (!((p == node) && (move == UP)))
     {
@@ -699,21 +758,29 @@ _asn1_expand_object_id (asn1_node node)
 		    {
 		      _asn1_str_cpy (name2, sizeof (name2), name_root);
 		      _asn1_str_cat (name2, sizeof (name2), ".");
-		      _asn1_str_cat (name2, sizeof (name2),
-				     (char *) p2->value);
+		      _asn1_str_cat (name2, sizeof (name2), (char *) p2->value);
 		      p3 = asn1_find_node (node, name2);
-		      if (!p3
-			  || (type_field (p3->type) != ASN1_ETYPE_OBJECT_ID)
-			  || !(p3->type & CONST_ASSIGN))
+		      if (!p3 || _asn1_is_up(p2, p3) ||
+			  (type_field (p3->type) != ASN1_ETYPE_OBJECT_ID) ||
+			  !(p3->type & CONST_ASSIGN))
 			return ASN1_ELEMENT_NOT_FOUND;
+
 		      _asn1_set_down (p, p2->right);
+		      if (p2->down)
+			_asn1_delete_structure (*list, &p2->down, 0);
+		      _asn1_delete_node_from_list(*list, p2);
 		      _asn1_remove_node (p2, 0);
 		      p2 = p;
 		      p4 = p3->down;
+		      max_constants = 0;
 		      while (p4)
 			{
 			  if (type_field (p4->type) == ASN1_ETYPE_CONSTANT)
 			    {
+			      max_constants++;
+			      if (max_constants == MAX_CONSTANTS)
+                                return ASN1_RECURSION;
+
 			      p5 =
 				_asn1_add_single_node (ASN1_ETYPE_CONSTANT);
 			      _asn1_set_name (p5, p4->name);
@@ -723,6 +790,8 @@ _asn1_expand_object_id (asn1_node node)
 			          if (tlen > 0)
 			            _asn1_set_value (p5, p4->value, tlen + 1);
 			        }
+			      _asn1_add_static_node2(list, p5);
+
 			      if (p2 == p)
 				{
 				  _asn1_set_right (p5, p->down);
@@ -738,6 +807,11 @@ _asn1_expand_object_id (asn1_node node)
 			  p4 = p4->right;
 			}
 		      move = DOWN;
+
+		      tries++;
+                      if (tries >= EXPAND_OBJECT_ID_MAX_RECURSION)
+                        return ASN1_RECURSION;
+
 		      continue;
 		    }
 		}
@@ -747,6 +821,7 @@ _asn1_expand_object_id (asn1_node node)
       else
 	move = RIGHT;
 
+      tries = 0;
       if (move == DOWN)
 	{
 	  if (p->down)
@@ -772,7 +847,6 @@ _asn1_expand_object_id (asn1_node node)
 	p = _asn1_find_up (p);
     }
 
-
   /*******************************/
   /*       expand DEFAULT        */
   /*******************************/
@@ -791,7 +865,8 @@ _asn1_expand_object_id (asn1_node node)
 		{
 		  _asn1_str_cpy (name2, sizeof (name2), name_root);
 		  _asn1_str_cat (name2, sizeof (name2), ".");
-		  _asn1_str_cat (name2, sizeof (name2), (char *) p2->value);
+		  if (p2->value)
+		    _asn1_str_cat (name2, sizeof (name2), (char *) p2->value);
 		  p3 = asn1_find_node (node, name2);
 		  if (!p3 || (type_field (p3->type) != ASN1_ETYPE_OBJECT_ID)
 		      || !(p3->type & CONST_ASSIGN))
@@ -935,9 +1010,9 @@ _asn1_type_set_config (asn1_node node)
 /*   otherwise ASN1_SUCCESS                                       */
 /******************************************************************/
 int
-_asn1_check_identifier (asn1_node node)
+_asn1_check_identifier (asn1_node_const node)
 {
-  asn1_node p, p2;
+  asn1_node_const p, p2;
   char name2[ASN1_MAX_NAME_SIZE * 2 + 2];
 
   if (node == NULL)
@@ -1016,7 +1091,7 @@ _asn1_check_identifier (asn1_node node)
 	p = p->right;
       else
 	{
-	  while (1)
+	  while (p)
 	    {
 	      p = _asn1_find_up (p);
 	      if (p == node)
